/*
	manifold  -  disperse  contents  of a file to one or more others.
	Copyright (C) 2002 R.L. Horn

	This program is free software; you  can  redistribute  it  and/or
	modify  it  under  the terms of the GNU General Public License as
	published by the Free Software Foundation; either  version  2  of
	the License, or (at your option) any later version.

	This  program  is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the  implied  warranty  of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to:

		The Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA
*/

/*
	Expand the following escaped characters:
		\a  bell
		\b  backspace
		\e  escape character
		\f  formfeed
		\n	newline
		\r	carriage return
		\t	horizontal tab
		\v  vertical tab
		\\  \ character
		\B	toggle bold
		\U	toggle underscore

		The following return information from the uname(2) function call
		and are the same as uname(1) options (but uppercase).

		\M  machine type
		\N  machine's network node hostname
		\R  operating system release
		\S  operating system name
		\V  current version level within operating system release
*/		

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#if STDC_HEADERS
#  include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#if (HAVE_UNAME && HAVE_SYS_UTSNAME_H)
#  include <sys/utsname.h>
#endif

#include "cookie.h"
#include "dsprsdef.h"

static int
c_strcpy(char **outp, char *s, int *count)
{
	unsigned int len;
	if (*count < 0) return -1;
	len = strlen(s);
	if (len > (unsigned) *count) len = *count;
	memcpy(*outp, s, len);
	(*count) -= len;
	(*outp) += len;
	return 0;
}

int
expand_cookies(char *s, int size, const char *format)
{
	char *outp = s, c;
	const char *inp = format;
	int escape = 0, count = size - 1;


#if HAVE_UNAME
	struct utsname u;
	uname(&u);
#endif

	if (count <= 0) return -1;
	count--;

	while ((c = *inp++) && (count)) {
		if (escape) {
			escape = 0;
			switch (c) {
				case 'n':
					*outp++ = '\n';
					count--;
					break;
				case 't':
					*outp++ = '\t';
					count--;
					break;
				case 'e':
					*outp++ = ESC;
					count--;
					break;
				case 'r':
					*outp++ = '\r';
					count--;
					break;
				case 'a':
					*outp++ = BELL;
					count--;
					break;
				case 'b':
					*outp++ = '\b';
					count--;
					break;
				case 'v':
					*outp++ = '\v';
					count--;
					break;
				case 'f':
					*outp++ = '\f';
					count--;
					break;
				case '\\':
					*outp++ = '\\';
					count--;
					break;

#if HAVE_UNAME
				case 'M':
					c_strcpy(&outp, u.machine, &count);
					break;
				case 'N':
					c_strcpy(&outp, u.nodename, &count);
					break;
				case 'R':
					c_strcpy(&outp, u.release, &count);
					break;
				case 'S':
					c_strcpy(&outp, u.sysname, &count);
					break;
				case 'V':
					c_strcpy(&outp, u.version, &count);
					break;
#elif HAVE_GETHOSTNAME
				case 'N': {
					char hostname[1024];
					if (!gethostname(hostname,1024)) {
						c_strcpy(&outp, hostname, &count);
					}
				}
				break;
#endif
				default:
					if ( count >= 2 ) {
						*outp++ = '\\';
						*outp++ = c;
						count -= 2;
					}
					break;
			}
		} else {
			escape = (c == '\\');
			if (!escape) {
				*outp++ = c;
				count--;
			}
		}
	}

	*outp = '\0';
	return size - count - 1;
}


int
expand_boldul(char *s, int size, const char *format)
{
	char *outp = s, c;
	const char *inp = format;
	int escape = 0, bold = 0, underscore = 0,
	    count = size - 1;


	if (count <= 0) return -1;
	count--;

	while ((c = *inp++) && (count)) {
		if (escape) {
			escape = 0;
			switch (c) {
				case 'B':
					bold = !bold;
					break;
				case 'U':
					underscore = !underscore;
					break;
				default:
					if ( count >= 2 ) {
						*outp++ = '\\';
						*outp++ = c;
						count -= 2;
					}
					break;
			}
		} else {
			escape = (c == '\\');
			if (!escape) {
				if ( underscore && (count >= 3) ) {
					*outp++ = '_';
					*outp++ = '\b';
					count -= 2;
				}
				*outp++ = c;
				count--;
				if ( bold && (count >= 2) ) {
					*outp++ = '\b';
					*outp++ = c;
					count -= 2;
				}
			}
		}
	}

	*outp = '\0';
	return size - count - 1;
}
