/*
	manifold  -  disperse  contents  of a file to one or more others.
	Copyright (C) 2002 R.L. Horn

	This program is free software; you  can  redistribute  it  and/or
	modify  it  under  the terms of the GNU General Public License as
	published by the Free Software Foundation; either  version  2  of
	the License, or (at your option) any later version.

	This  program  is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the  implied  warranty  of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to:

		The Free Software Foundation, Inc.
		59 Temple Place, Suite 330
		Boston, MA  02111-1307
		USA
*/

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <errno.h>

#if STDC_HEADERS
#  if TM_IN_SYS_TIME
#    if TIME_WITH_SYS_TIME
#      include <time.h>
#      include <sys/time.h>
#    else
#      include <sys/time.h>
#    endif
#  else
#    include <time.h>
#  endif	
#  include <string.h>
#endif

#if HAVE_UNISTD_H
#  include <unistd.h>
#else
#  if !HAVE_GETOPT
     int getopt(int argc, char *argv[], char *opts);
     extern char *optarg;
     extern int optind, opterr, optopt;
#  endif
#endif

#if HAVE_GETOPT_LONG && HAVE_GETOPT_H
#  include <getopt.h>
#endif

#include <signal.h>
#include <ctype.h>

#if HAVE_LIBZ
#  if HAVE_ZLIB_H
#    include <zlib.h>
#  else
#    warning "!!! zlib.h not found...attempting my own prototypes !!!"
     typedef void * gzFile;
     extern gzFile gzopen(const char *path, const char *mode);
     extern int gzprintf(gzFile file, const char *format, ...);
     extern int gzputc(gzFile file, int c);
     extern int gzclose(gzFile file);
#  endif
#endif

#include "usage.h"
#include "cookie.h"
#include "dsprsdef.h"
#include "pferror.h"

#if USE_TERMINFO
#  include "dterm.h"
#endif

/*
	manifold		take input from a single file and disperse to
					several different files
*/

/* exit codes */
#define EXIT_BAD_OPTION 1
#define EXIT_INFILE_OPEN_ERROR 2
#define EXIT_OUTFILE_OPEN_ERROR 3
#define EXIT_WRITE_ERROR 4

#ifndef NULL
#define NULL ((void *) 0)
#endif

#if HAVE_GETOPT_LONG
#  define GETOPT(argc,argv,optstring,longopts,longindex) \
	getopt_long(argc,argv,optstring,longopts,longindex)
#else
#  define GETOPT(argc,argv,optstring,longopts,longindex) \
	getopt(argc,argv,optstring)
#endif

#if SETVBUF_REVERSED
#  define SETVBUF(stream,buf,mode,size) setvbuf(stream,mode,buf,size)
#else
#  define SETVBUF setvbuf
#endif

typedef struct {
	FILE *stream;
#if HAVE_LIBZ
	gzFile gzstream;
#endif
	char *name;
} output_struct;

char *pgmname;
int opt_append = 0,
	opt_header = 0, opt_footer = 0, opt_utc = 0,
	opt_linebuffered = 0,
	opt_show_nonprinting = 0,
	opt_using_terminfo = 0,
	opt_ignore_interrupt = 0,
	opt_quiet = 0,
	tee;

#if USE_TERMINFO
  static char *term = NULL;
#endif

static void
write_hdrftr(output_struct outfile[], int ofcount, const char *timefmt)
{
	int j;
	time_t t = time(NULL);
	char ts[HF_SIZE], cs[HF_SIZE];

	/* get the time */
	if (opt_utc)
		strftime(ts,HF_SIZE,timefmt,gmtime(&t));
	else
		strftime(ts,HF_SIZE,timefmt,localtime(&t));

	/* now expand our own escaped sequences (except for bold and underlined) */
	expand_cookies(cs,HF_SIZE,ts);
	/* expand bold and underlined for dumb terminals */
	expand_boldul(ts,HF_SIZE,cs);

	for (j = 0; j < ofcount; j++) {
#		if USE_TERMINFO
			if ( (outfile[j].stream == stdout) && opt_using_terminfo )
			{
				if (dterm_print(term,cs)) {
					/* curses failure...try doing it the ordinary way  */
					if (!fprintf(outfile[j].stream,"%s",ts)) {
						print_error(errno,outfile[j].name);
						fclose(outfile[j].stream);
						outfile[j].stream = NULL;
					}
				}
			} else {
#		else
		{
#endif
#if HAVE_LIBZ
			if (outfile[j].gzstream) {
				if (!gzprintf(outfile[j].gzstream,"%s",ts)) {
					print_error(errno,outfile[j].name);
					gzclose(outfile[j].gzstream);
					outfile[j].gzstream = NULL;
				}
			} else
#endif
			if (outfile[j].stream) {
				if (!fprintf(outfile[j].stream,"%s",ts)) {
					print_error(errno,outfile[j].name);
					fclose(outfile[j].stream);
					outfile[j].stream = NULL;
				}
			}
		}
	}
}

#if HAVE_LIBZ
static int
is_gz(const char *s)
{
	int len = strlen(s);
	if (len < 3) return 0;
	return (strcmp(s+len-3,".gz") == 0);
}
#endif

int
main(int argc, char *argv[])
{
	FILE *infile;
	output_struct outfile[MAX_FILES];
	int c,j,ofcount, output_error,
		argidx,
		stdout_used = 0, stderr_used = 0;
	char octal[8],
		 *header_format = DEFAULT_HEADER,
		 *footer_format = DEFAULT_FOOTER;

#if HAVE_GETOPT_LONG
	static struct option long_options[] = {
		{"help",0,0,'@'},
		{"version",0,0,'V'},
		{"append",0,0,'a'},
		{"quiet",0,0,'q'},
		{"ignore-interrupts",0,0,'i'},
		{"linebuffer",0,0,'l'},
		{"header",0,0,'h'},
		{"header-string",1,0,'H'},
		{"footer",0,0,'f'},
		{"footer-string",1,0,'F'},
		{"utc",0,0,'u'},
		{"show-nonprinting",0,0,'v'},
		{"use-terminfo",0,0,'t'},
		{"terminal",1,0,'T'},
		{0,0,0,0}
	};

	int option_index = 0;
#endif

	if ( (pgmname = strrchr(argv[0],'/')) ) {
		pgmname++;
	} else {
		pgmname = argv[0];
	}

	tee = !strncmp(pgmname,"tee",3);

	/* process any command line options */

#ifndef HAVE_GETOPT_LONG
	/* we don't actually use long options, but go through the
	   argv's and check for "--help" and "--version" */

	for (j = 1; j < argc; j++) {
		if (!strcmp(argv[j],"--help")) {
			usage(stdout);
			exit(0);
		}

		if (!strcmp(argv[j],"--version")) {
			version();
			exit(0);
		}
	}
#endif

	while ((c = GETOPT(argc,argv,"VF:H:T:afhilqtuv",long_options,&option_index)) != EOF) {
		switch (c) {
			case 'V':
				version();
				exit(0);
			case '@':
				usage(stdout);
				exit(0);
			case 'a':
				opt_append = 1;
				break;
			case 'q':
				opt_quiet = 1;
				break;
			case 'l':
				opt_linebuffered = 1;
				break;
			case 'h':
				opt_header = 1;
				break;
			case 'H':
				opt_header = 1;
				header_format = optarg;
				break;
			case 'f':
				opt_footer = 1;
				break;
			case 'F':
				opt_footer = 1;
				footer_format = optarg;
				break;
			case 'v':
				opt_show_nonprinting = 1;
				break;
			case 'u':
				opt_utc = 1;
				break;
			case 'i':
				opt_ignore_interrupt = 1;
				break;
#			if USE_TERMINFO
				case 't':
#					if HAVE_ISATTY
						opt_using_terminfo = isatty(fileno(stdout));
#					else
						opt_using_terminfo = 1;
#					endif
					break;
				case 'T':
					opt_using_terminfo = 1;
					term = optarg;
					break;
#			else
				/* Fail silently for compatibility's sake */
				case 't': break;
				case 'T': break;
#			endif

			default:
				fputs("Try `" PACKAGE " --help' for more information.\n",stderr);
				exit(EXIT_BAD_OPTION);
		} /* switch (c) */
	} /* while */

	argidx = optind;

	if (opt_ignore_interrupt) {
#		ifdef _POSIX_SOURCE
			struct sigaction sigact;

			sigact.sa_handler = SIG_IGN;
			sigemptyset (&sigact.sa_mask);
			sigact.sa_flags = 0;
			sigaction (SIGINT, &sigact, NULL);
#		else                           /* !_POSIX_SOURCE */
			signal (SIGINT, SIG_IGN);
#endif                          /* _POSIX_SOURCE */
	}

	/* Don't let us be killed if one of the output files is a pipe that
	   doesn't consume all its input.  */

#	ifdef _POSIX_SOURCE
		{
			struct sigaction sigact;

			sigact.sa_handler = SIG_IGN;
			sigemptyset (&sigact.sa_mask);
			sigact.sa_flags = 0;
			sigaction (SIGPIPE, &sigact, NULL);
		}
#	else
		signal (SIGPIPE, SIG_IGN);
#	endif

	if ( (argc <= argidx) || tee ) {
		/* no file(s) specified at all or we're tee(1) */
		infile = stdin;
	} else {
		if (!strcmp(argv[optind],"-")) {
			/* stdin */
			infile = stdin;
		} else {
			infile = fopen(argv[optind],"rb");
			if (!infile) {
				print_error(errno,argv[optind]);
				exit(EXIT_INFILE_OPEN_ERROR);
			}
		}
		argidx++;
	}

	if (argc <= (argidx)) {
		outfile[0].stream = stdout;
#if HAVE_LIBZ
		outfile[0].gzstream = NULL;
#endif
		outfile[0].name = "stdout";
		ofcount = 1;
	} else {
		ofcount = 0;
		if (tee) {
			outfile[ofcount].stream = stdout;
#if HAVE_LIBZ
			outfile[ofcount].gzstream = NULL;
#endif
			outfile[ofcount].name = "stdout";
			stdout_used = 1;
			ofcount++;
		}

		for (j = argidx; j < argc; j++) {
			int append_thisfile = opt_append;
			char *arg = argv[j];

			if (!strncmp(arg,">>",2) && !opt_append) {
				append_thisfile = 1;
				arg += 2;
			}

			if ((!strcmp(arg,"&1") || !strcmp(arg,"-"))) { 
				/* stdout */
				if (!stdout_used) {
					outfile[ofcount].stream = stdout;
#if HAVE_LIBZ
					outfile[ofcount].gzstream = NULL;
#endif
					outfile[ofcount].name = "stdout";
					stdout_used = 1;
					ofcount++;
				}
			} else if (!strcmp(arg,"&2") && !stderr_used) {
				/* stderr */
				outfile[ofcount].stream = stderr;
#if HAVE_LIBZ
				outfile[ofcount].gzstream = NULL;
#endif
				outfile[ofcount].name = "stderr";
				stderr_used = 1;
				opt_quiet = 1;
				ofcount++;
			} else {
				/* regular filename, open for append */
				void *retp;
				char *mode = append_thisfile ? "ab":"wb";
#if HAVE_LIBZ
				if (is_gz(arg)) {
					retp = outfile[ofcount].gzstream = gzopen(arg,mode);
					outfile[ofcount].stream = NULL;
				} else {
					retp = outfile[ofcount].stream = fopen(arg,mode);
					outfile[ofcount].gzstream = NULL;
				}
#else
				retp = outfile[ofcount].stream = fopen(arg,mode);
#endif
				if (!retp) {
					print_error(errno,arg);
				} else {
					outfile[ofcount].name = arg;
					if (outfile[ofcount].stream && opt_linebuffered)
						SETVBUF(outfile[ofcount].stream,NULL,_IOLBF,0);
					ofcount++;
				}
			}
			if (ofcount >= MAX_FILES) break;
		}

		if (ofcount > MAX_FILES) {
			/* !!!!!!!!!  Oh well... */
			ofcount = MAX_FILES;
		}
	}

	/*
		we now have a list of output file pointers, all that remains
		is to disperse the input to the various outputs
	*/

	/* write header if requested */
	if (opt_header) write_hdrftr(outfile,ofcount,header_format);

	while ((c = fgetc(infile)) != EOF) {
		if (opt_show_nonprinting) {
			if (!isprint(c) && (c != '\n') && (c != '\t'))
				sprintf(octal,"%03o",c);
			else
				octal[0] = '\0';
		}
		
		for (j = 0; j < ofcount; j++) {
#if HAVE_LIBZ
			if (outfile[j].gzstream) {
				if (opt_show_nonprinting && octal[0])
					output_error = (!gzprintf(outfile[j].gzstream,"\\%s",octal));
				else
					output_error = (gzputc(outfile[j].gzstream,c) == EOF);
				if (output_error) {
					print_error(errno,outfile[j].name);
					gzclose(outfile[j].gzstream);
					outfile[j].gzstream = NULL;
				}
			} else
#endif
			if (outfile[j].stream) {
				if (opt_show_nonprinting && octal[0])
					output_error = (!fprintf(outfile[j].stream,"\\%s",octal));
				else
					output_error = (fputc(c,outfile[j].stream) == EOF);
				if (output_error) {
					print_error(errno,outfile[j].name);
					fclose(outfile[j].stream);
					outfile[j].stream = NULL;
				}
			}
		}
	}

	/* close the input file */
	fclose(infile);

	/* write footer if requested */
	if (opt_footer) write_hdrftr(outfile,ofcount,footer_format);

	/* close the output files */
	for (j = 0; j < ofcount; j++) {
#if HAVE_LIBZ
		if (outfile[j].gzstream)
			gzclose(outfile[j].gzstream);
		else
#endif
		if (outfile[j].stream)
			fclose(outfile[j].stream);
	}

	return 0;
}
